%--------------------------------------------------------------------------------
% Project       : LPA-Speech
% File          : LPA_Speech.pl
% Purpose       : Prolog Interface to SAPI speech recognition and synthesis engine.
% System        : LPA-Prolog 4.300, MS SAPI 5.1
% Author        : Arlo Lyle
% Institution   : The University of Georgia Artificial Intelligence Center
% Last Modified : 5/6/06
%--------------------------------------------------------------------------------

%--------------------------------------------------------------------------------
:- dynamic(dll_loaded/0).
:- dynamic(sapi_installed/0).
:- dynamic(attribute/3).
:- dynamic(volume/1).
:- dynamic(rate/1).
%--------------------------------------------------------------------------------

attribute(gender,"Gender=Male",required).
attribute(age,"Age=Adult",required).
volume(100).
rate(0).

%--------------------------------------------------------------------------------
% ensure_foreign_library_loaded
%  Ensures that lpa_speech.dll is loaded into the system. lpa_speech.dll must be
%  loaded before any calls may be made to the predicates it defines.
%  If lpa_speech.dll is not loaded, it is then loaded.
%--------------------------------------------------------------------------------
ensure_dll_loaded :- dll_loaded.

ensure_dll_loaded :- 
   winapi((kernel32,'LoadLibraryA'),[`lpa_speech.dll`],0,_),
   winapi((kernel32,'LoadLibraryA'),[`registry.dll`],0,_),
   assert(dll_loaded).
  
%--------------------------------------------------------------------------------
% listen(-Sentence)
%  Uses SAPI's speech recognition engine to listen for voice input. Voice input
%  is returned in the form of a compact string (eg. `hello`).
%--------------------------------------------------------------------------------
listen(Sentence) :-
   sapi_installed,
   ensure_dll_loaded,
   winapi((lpa_speech,lpa_listen),[],0,Result),
   wintxt(Result,0,0,Sentence).

%--------------------------------------------------------------------------------
% speak(+Sentence)
%  Uses SAPI's speech sysnthesis engine to speak a line of text. This method
%  uses whatever the current voice settings are as set by the user. Input to this
%  method can be in the form of a list of characters, compact string, or atom.
%--------------------------------------------------------------------------------
speak(Sentence) :-
   atom(Sentence), !,
   name(Sentence,Sentence2),
   speak(Sentence2).

speak(Sentence) :-
   list(Sentence), !,
   string_chars(Sentence2,Sentence),
   speak(Sentence2).

speak(Sentence) :-
   sapi_installed,
   ensure_dll_loaded,
   winapi((lpa_speech,lpa_speak_no_attr),[Sentence],0,_).

%--------------------------------------------------------------------------------
% speak_with_attributes(+Sentence)
%  Uses SAPI's speech synthesis engine to speak a line of text. This method
%  uses the voice attributes that are set by the following predicates,
%  set_gender/2, set_age/2, set_rate/1, and set_volume/1. The gender and age
%  attributes are used in the following manner. The windows registry is searched
%  for voices that match the settings of gender and age. Of all the voices that
%  match, the best match is used to speak the line of text. Attributes marked as
%  required must be present in any voice selected by the search. Optional
%  attributes merely influence which voice is chosen as the best match. Input to
%  this method can be in the form of a list of characters, compact string, or atom.
%--------------------------------------------------------------------------------
speak_with_attributes(Sentence) :-
   atom(Sentence), !,
   name(Sentence,Sentence2),
   speak_with_attributes(Sentence2).

speak_with_attributes(Sentence) :-
   list(Sentence), !,
   string_chars(Sentence2,Sentence),
   speak_with_attributes(Sentence2).

speak_with_attributes(Sentence) :-
   sapi_installed,
   ensure_dll_loaded,
   volume(Volume),
   rate(Rate),
   get_required_list(Output1),
   string_chars(RequiredList,Output1),
   get_optional_list(Output2),
   string_chars(OptionalList,Output2),
   winapi((lpa_speech,lpa_speak),[Sentence,Rate,Volume,RequiredList,OptionalList],0,_).

get_required_list([]) :-
   \+ attribute(_,_,required),
   !.

get_required_list(RequiredList) :-
   findall(Attribute,
	   (
	     attribute(_,Attribute,required)
	   ),
	  [FirstAttribute|Rest]),
   append_list(FirstAttribute,Rest,RequiredList).

get_optional_list([]) :-
   \+ attribute(_,_,optional),
   !.

get_optional_list(OptionalList) :-
   findall(Attribute,
	   (
	     attribute(_,Attribute,optional)
	   ),
	  [FirstAttribute|Rest]),
   append_list(FirstAttribute,Rest,OptionalList).

append_list(CompletedList,[],CompletedList).

append_list(PartialList,[List|RestOfLists],CompletedList) :-
   append(PartialList,";",Temp),
   append(Temp,List,NewPartialList),
   append_list(NewPartialList,RestOfLists,CompletedList).

%--------------------------------------------------------------------------------
% set_gender(+Gender,+Condition)
%  This method sets the gender attribute for use with speak_with_attributes/1.
%  Gender can be set to either male or female. The condition can be either set to
%  required or optional depending on whether the gender is required or just
%  preferred. The default setting for gender is male and required.
%--------------------------------------------------------------------------------
set_gender(male,optional) :-
   retract(attribute(gender,_,_)),
   assert(attribute(gender,"Gender=Male",optional)).

set_gender(male,required) :-
   retract(attribute(gender,_,_)),
   assert(attribute(gender,"Gender=Male",required)).

set_gender(female,optional) :-
   retract(attribute(gender,_,_)),
   assert(attribute(gender,"Gender=Female",optional)).

set_gender(female,required) :-
   retract(attribute(gender,_,_)),
   assert(attribute(gender,"Gender=Female",required)).

%--------------------------------------------------------------------------------
% set_age(+Age,+Condition)
%  This method sets the age attribute for use with speak_with_attributes/1.
%  Age can be set to the following values child, teen, adult, and senior. The
%  condition can be either set to required or optional depending on whether the
%  age is required or just preferred. The default setting for age is adult and
%  required.
%--------------------------------------------------------------------------------
set_age(child,optional) :-
   retract(attribute(age,_,_)),
   assert(attribute(age,"Age=Child",optional)).

set_age(child,required) :-
   retract(attribute(age,_,_)),
   assert(attribute(age,"Age=Child",required)).

set_age(teen,optional) :-
   retract(attribute(age,_,_)),
   assert(attribute(age,"Age=Teen",optional)).

set_age(teen,required) :-
   retract(attribute(age,_,_)),
   assert(attribute(age,"Age=Teen",required)).

set_age(adult,optional) :-
   retract(attribute(age,_,_)),
   assert(attribute(age,"Age=Adult",optional)).

set_age(adult,required) :-
   retract(attribute(age,_,_)),
   assert(attribute(age,"Age=Adult",required)).

set_age(senior,optional) :-
   retract(attribute(age,_,_)),
   assert(attribute(age,"Age=Senior",optional)).

set_age(senior,required) :-
   retract(attribute(age,_,_)),
   assert(attribute(age,"Age=Senior",required)).

%--------------------------------------------------------------------------------
% set_rate(+Rate)
%  This method sets the rate of speech for use with speak_with_attributes/1.
%  The rate value must be a number between -10 and 10, with -10 being the slowest
%  and 10 being the fastest. The default setting for rate is 0.
%--------------------------------------------------------------------------------
set_rate(Rate) :-
   retract(rate(_)),
   assert(rate(Rate)).

%--------------------------------------------------------------------------------
% set_volume(+Volume)
%  This method sets the volume of speech for use with speak_with_attributes/1.
%  The volume value must be a number between 0 and 100, with 0 being the lowest
%  and 100 being the highest. The default setting for volume is 100.
%--------------------------------------------------------------------------------
set_volume(Volume) :-
   retract(volume(_)),
   assert(volume(Volume)).
   
%--------------------------------------------------------------------------------
% ensure_SAPI_installed
%  This method checks to make sure SAPI is installed
%--------------------------------------------------------------------------------
ensure_SAPI_installed :-
   ensure_dll_loaded,
   fcreate(buffer,[],-2,256,0),
   winapi((registry,'GetNonNEDValue'),[`HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Speech\Recognizers\Tokens\SAPI5SampleEngine\Attributes\`,
   	`Vendor`,buffer],0,Result),
   Result = 1,
   wintxt(buffer,0,0,`Microsoft`),
   fclose(buffer),
   assert(sapi_installed).   
      
:- ensure_SAPI_installed.
